from PySide6.QtWidgets import QLineEdit, QWidget

from ..objects import ContextObj, SizedObj, EventedObj
from ..types import Applicable


class LineInput(ContextObj, QLineEdit):
    def __init__(self, parent: QWidget, name: str, visible: bool = True):
        QLineEdit.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)

    async def init(
            self, *,
            placeholder: str = '', text: str = '', hidden: bool = False,
            sizes: Applicable = SizedObj.Sizes(), events: Applicable = EventedObj.Events()
    ) -> 'LineInput':
        self.setText(text)
        self.setPlaceholderText(placeholder)
        if hidden:
            self.hide_echo()
        return await sizes(await events(self))

    def hide_echo(self):
        self.setEchoMode(QLineEdit.EchoMode.Password)

    def show_echo(self):
        self.setEchoMode(QLineEdit.EchoMode.Normal)

    def toggle_echo(self):
        self.hide_echo() if self.echoMode() == QLineEdit.EchoMode.Normal else self.show_echo()



import contextlib
import threading
import uvicorn
import asyncio
import time
from uvicorn import Config  # to import while using `Server`


class Server(uvicorn.Server):
    def install_signal_handlers(self):  # overriding to disable
        pass

    @contextlib.contextmanager
    def run_in_thread(self):
        self.thread = threading.Thread(target=self.run)
        self.thread.start()
        try:
            while not self.started:
                time.sleep(1e-3)
            yield
        finally:
            self.stop()
        return self

    def stop(self):
        self.should_exit = True
        self.thread.join()
        asyncio.get_event_loop().stop()



from PySide6.QtCore import Qt


class SideMenuExt:
    def __init__(self, expand_to, expand_orientation: Qt.Orientation = Qt.Horizontal):
        self.expand_to = expand_to
        self.expand_orientation = expand_orientation

    def __dimension(self):
        return self.width() if self.expand_orientation is Qt.Horizontal else self.height()

    def expand(self, expand_to: int = None):
        if expand_to:
            self.expand_to = expand_to
        self.setFixedWidth(self.expand_to)

    def shrink(self) -> None:
        self.setFixedWidth(0)

    def toggle(self) -> None:
        self.expand() if self.__dimension() == 0 else self.shrink()



from PySide6.QtCore import QPropertyAnimation, Property
from PySide6.QtWidgets import QWidget
from time import sleep

from ...misc import ConditionalThreadQueue
from ..label import Label
from ...qasyncio import asyncSlot


class ErrorLabel(Label):
    def __init__(self, parent: QWidget, name: str = None, visible: bool = True):
        super().__init__(parent, name if name else self.__class__.__name__, visible)
        self.__ctq = ConditionalThreadQueue()
        self.__opacity: float = 1
        self.__duration: float = 0.5

    async def init(self, *args, **kwargs) -> 'ErrorLabel':
        await super().init(*args, **kwargs)
        return self

    def setText(self, text: str, delay: float = 2, duration: int = 0.5) -> None:
        if delay == 0:  # if delay is 0, just set the text
            return Label.setText(self, '')

        if not text:  # text == '' means instantly clear the text and hide label without `post` action
            self.setVisible(False)
            return Label.setText(self, '')

        def pre():
            self._set_opacity(1)
            Label.setText(self, text)
            self.setVisible(True)
            sleep(delay)

        def post():
            Label._emit(self.reduce)

        self.__duration = duration
        self.__ctq.new(pre, post)

    @asyncSlot()
    async def reduce(self):
        self.animation = QPropertyAnimation(self, b"_opacity")
        self.animation.setDuration(int(self.__duration * 1000))
        self.animation.setStartValue(1.0)
        self.animation.setEndValue(0.0)
        self.animation.start()

    def _get_opacity(self):
        return self.__opacity

    def _set_opacity(self, opacity):
        self.__opacity = opacity
        self.setStyleSheet(f'color: rgba(255, 0, 0, {opacity});')

    _opacity = Property(float, _get_opacity, _set_opacity)



from PySide6.QtWidgets import QSplitter
from loguru import logger as _logger

from ...enums import Orientation, SizePolicy


class SplitterWidgetExt(Orientation):
    splitter: QSplitter

    def __init__(self, expand_to: int,
                 expand_min: int = None, expand_max: int = None, orientation: Orientation.Orientation = None,
                 policy: tuple[SizePolicy.SizePolicy, SizePolicy.SizePolicy] = (SizePolicy.Expanding, SizePolicy.Expanding)):
        self.expand_to: int = expand_to
        self.setSizePolicy(*policy)

        if expand_min or expand_max:
            if orientation:
                dimension = 'width' if orientation is Orientation.Horizontal else 'height'
                if expand_min:
                    getattr(self, f'setMinimum{dimension.capitalize()}')(expand_min)
                if expand_max:
                    getattr(self, f'setMaximum{dimension.capitalize()}')(expand_max)
            else:
                _logger.info('`expand_min` and `expand_max` set but `orientation` is `None`')

    def _index(self) -> int:
        for index in range(self.splitter.count()):
            if self == self.splitter.widget(index):
                return index
        raise IndexError(f'{self} not found in {self.splitter} widgets')

    def _set_size(self, size: int):
        index = self._index()
        self.splitter.setSizes([*self.splitter.sizes()[:index], size, *self.splitter.sizes()[index + 1:]])

    def expand(self, size: int = None):
        if not self.splitter.sizes()[self._index()]:
            self._set_size(size if size else self.expand_to)

    def shrink(self, size: int = None) -> None:
        self._set_size(size if size else 0)

    def toggle(self, shrink_to: int = None, expand_to: int = None) -> None:
        if self.splitter.sizes()[self._index()]:
            self.shrink(shrink_to)
        else:
            self.expand(expand_to)



from PySide6.QtWidgets import QWidget
from PySide6.QtGui import QIcon
from typing import Callable

from ..button import Button
from ...types import Icon
from ...qasyncio import asyncSlot


class FavouriteButton(Button):
    def __init__(self, parent: QWidget, name: str = None, visible: bool = True, is_favourite: bool = False):
        super().__init__(parent, name if name else self.__class__.__name__, visible)
        self.is_favourite: bool = is_favourite
        self.if_set_icon: Icon = Icon('star-fill.svg', (30, 30))
        self.if_unset_icon: Icon = Icon('star.svg', (30, 30))

    async def init(
            self, *,
            if_set_icon: Icon = None, if_unset_icon: Icon = None, pre_slot: Callable[..., bool],
            **kwargs
    ) -> 'Button':
        await super().init(events=Button.Events(on_click=lambda: self.slot(pre_slot)), **kwargs)
        if if_unset_icon:
            self.setIcon(QIcon(if_unset_icon.icon))
            self.setIconSize(if_unset_icon.size)
        if if_set_icon:
            self.if_set_icon = if_set_icon
        if if_unset_icon:
            self.if_unset_icon = if_unset_icon
        return self

    @asyncSlot()
    async def slot(self, pre_slot: Callable[..., bool]):
        # toggle to change state
        self.set(not self.is_favourite)
        if not await pre_slot():
            # toggle once more to come back to prev state if `pre_slot` returns `False`
            self.set(not self.is_favourite)

    def set_favourite(self):
        self.setIcon(self.if_set_icon.icon)
        self.setIconSize(self.if_set_icon.size)
        self.is_favourite = True

    def unset_favourite(self):
        self.setIcon(self.if_unset_icon.icon)
        self.setIconSize(self.if_unset_icon.size)
        self.is_favourite = False

    def set(self, is_favourite: bool):
        self.is_favourite = is_favourite
        if self.is_favourite:
            self.set_favourite()
        else:
            self.unset_favourite()



from PySide6.QtGui import QPixmap, QIcon, QImage
from PySide6.QtCore import QSize, QBuffer, QIODevice
# from loguru import logger
import base64
import os

from .size import Size


class Icon:
    IconType = type[QIcon, str, bytes]
    SizeType = type[Size, QSize, tuple[int, int], Ellipsis, None]


    root = os.path.join(os.path.dirname(__file__), os.path.pardir, '.assets', 'icons')
    __icon: QIcon = None
    __size: QSize = None

    def __init__(self, instance: IconType, size: SizeType = None):
        self.icon = instance
        if size:
            self.size = size

    @property
    def icon(self) -> QIcon:
        return self.__icon

    @property
    def size(self) -> QSize:
        return self.__size

    @icon.setter
    def icon(self, instance: IconType) -> None:
        """

        :param instance: QIcon, string filename with already set Icon.root, icon bytes
        :return:
        """
        def from_bytes(icon_bytes: bytes):
            pixmap = QPixmap()
            png = base64.b64encode(icon_bytes).decode('utf-8')
            pixmap.loadFromData(base64.b64decode(png))
            self.__icon = QIcon(pixmap)

        if isinstance(instance, QIcon):  # directly set
            self.__icon = instance
        elif isinstance(instance, str):  # might be `paths` or `str(bytes)`
            if os.path.exists(filepath := os.path.join(os.path.abspath(self.root), instance)):
                self.__icon = QIcon(filepath)
            else:
                try:  # handle `eval(intance)` for case if instance is invalid `path`
                    if isinstance((icon_bytes := eval(instance)), bytes):
                        from_bytes(icon_bytes)
                except Exception:
                    self.__icon = QIcon()
                    # logger.warning(f'{instance=} is invalid path')
        elif isinstance(instance, bytes):
            from_bytes(instance)
        else:
            self.__icon = QIcon()
            # logger.warning(f'unknown type ({type(instance)}) of instance ({instance}), can not set Icon.icon')
        self.icon.addPixmap(self.__icon.pixmap(1000), mode=QIcon.Disabled)  # fix of `QPushButton.setDisabled()`

    @size.setter
    def size(self, size: SizeType) -> None:
        """

        :param size: QSize, qcontextapi.Size, tuple[int, int]
        :return:
        """
        if isinstance(size, tuple):
            self.__size = QSize(*size)
        elif isinstance(size, Size):
            self.__size = QSize(*size.size)
        elif isinstance(size, QSize):
            self.__size = size
        else:
            raise AttributeError(f'unknown type ({type(size)}) of size ({size}), can not set Icon.size')

    def adjusted(self, icon: IconType = None, size: SizeType = None) -> 'Icon':
        # if not icon and not size:
        #     logger.warning('Icon.adjusted() called without parameters. Redundant call.')
        if icon:
            self.icon = icon
        if size:
            self.size = size
        return self

    @staticmethod
    def bytes(icon: IconType, size: SizeType = None) -> bytes:
        icon = Icon(icon, size)
        # image = QImage(icon.icon.pixmap(icon.size).toImage())
        image = QImage(icon.icon.pixmap(QSize(512, 512)).toImage())
        buffer = QBuffer()
        buffer.open(QIODevice.WriteOnly)
        image.save(buffer, 'JPG')
        image_bytes = buffer.data()
        buffer.close()
        return image_bytes



from PySide6.QtWidgets import QPushButton, QWidget
from PySide6.QtGui import QIcon

from ..objects import ContextObj, SizedObj, EventedObj
from ..types import Applicable, Icon


class Button(ContextObj, QPushButton):
    def __init__(self, parent: QWidget, name: str, visible: bool = True):
        QPushButton.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)

    async def init(
            self, *,
            text: str = '', icon: Icon = None, disabled: bool = False,
            sizes: Applicable = SizedObj.Sizes(), events: Applicable = EventedObj.Events()
    ) -> 'Button':
        self.setText(text)
        self.setDisabled(disabled)
        if icon:
            self.setIcon(QIcon(icon.icon))
            self.setIconSize(icon.size)
        return await sizes(await events(self))



from PySide6.QtCore import QObject
from loguru import logger

from ..enums import SizePolicy, Alignment
from ..types import Applicable, Size


class SizedObj(SizePolicy):
    @staticmethod
    def Sizes(
            *,
            policy: tuple[SizePolicy, Size] = None,
            vpolicy: SizePolicy = None,
            hpolicy: SizePolicy = None,

            # margins: ? = ?,
            # padding: ? = ?,
            alignment: Alignment.Alignment = None,

            size: Size = None,
            width: int = None,
            height: int = None,

            min_size: Size = None,
            min_width: int = None,
            min_height: int = None,

            max_size: Size = None,
            max_width: int = None,
            max_height: int = None,

            fixed_size: Size = None,
            fixed_width: int = None,
            fixed_height: int = None
    ) -> Applicable:
        async def apply(self: QObject):
            nonlocal policy, vpolicy, hpolicy
            if policy:
                self.setSizePolicy(*policy)
                if vpolicy or hpolicy:
                    SizedObj._warning(self)
            else:
                policy_inner = self.sizePolicy()
                if vpolicy:
                    policy_inner.setVerticalPolicy(vpolicy)
                elif hpolicy:
                    policy_inner.setHorizontalPolicy(hpolicy)
                self.setSizePolicy(policy_inner)

            if alignment:
                self.setAlignment(alignment)

            if size:
                self.resize(*size.size)
                if width or height:
                    SizedObj._warning(self)
            elif width:
                self.resize(width, self.height())
            elif height:
                self.resize(self.width(), height)

            if min_size:
                self.setMinimumSize(*min_size.size)
                if min_width or min_height:
                    SizedObj._warning(self)
            elif min_width:
                self.setMinimumWidth(min_width)
            elif min_height:
                self.setMinimumHeight(min_height)

            if max_size:
                self.setMaximumSize(*max_size.size)
                if max_width or max_height:
                    SizedObj._warning(self)
            elif max_width:
                self.setMaximumWidth(max_width)
            elif max_height:
                self.setMaximumHeight(max_height)

            if fixed_size:
                self.setFixedSize(*fixed_size.size)
                if fixed_width or fixed_height:
                    SizedObj._warning(self)
            elif fixed_width:
                self.setFixedWidth(fixed_width)
            elif fixed_height:
                self.setFixedHeight(fixed_height)
            return self
        return apply

    @staticmethod
    def _warning(handler: QObject) -> None:
        logger.warning(f'Review `{handler.objectName()}.sized()` arguments')




from .favourite_button import FavouriteButton
from .image_button import ImageButton
from .menu_button import MenuButton
from .search_bar import SearchBar
from .date_picker import DateTimePicker
from .error_label import ErrorLabel



from PySide6.QtWidgets import QVBoxLayout, QWidget, QHBoxLayout
from PySide6.QtCore import Qt
from typing import Union
from abc import ABC

from .extensions import LayoutExt
from ..objects import ContextObj
from ..enums import Alignment, SizePolicy, Orientation


class Layout(ABC, Alignment, SizePolicy, Orientation):
    @classmethod
    def horizontal(cls, parent: QWidget = None, name: str = None) -> 'HLayout':
        return HLayout(parent, name)

    @classmethod
    def vertical(cls, parent: QWidget = None, name: str = None) -> 'VLayout':
        return VLayout(parent, name)

    @classmethod
    def oriented(cls, orientation: Qt.Orientation, parent: QWidget = None, name: str = None) -> Union['VLayout', 'HLayout']:
        return Layout.vertical(parent, name) if orientation is Layout.Vertical else Layout.horizontal(parent, name)


class VLayout(ContextObj, LayoutExt, QVBoxLayout):
    def __init__(self, parent: QWidget = None, name: str = None):
        QVBoxLayout.__init__(self, parent)
        ContextObj.__init__(self, parent, name, True)


class HLayout(ContextObj, LayoutExt, QHBoxLayout):
    def __init__(self, parent: QWidget = None, name: str = None):
        QHBoxLayout.__init__(self, parent)
        ContextObj.__init__(self, parent, name, True)



from PySide6.QtWidgets import QFrame, QWidget, QLayout

from ..objects import ContextObj, SizedObj, EventedObj
from ..types import Applicable


class Frame(ContextObj, QFrame):
    def __init__(self, parent: QWidget, name: str, visible: bool = True, stylesheet: str = ''):
        QFrame.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)
        self.setStyleSheet(stylesheet)

    async def init(
            self, *,
            style: ... = None, layout: QLayout = None,
            sizes: Applicable = SizedObj.Sizes(), events: Applicable = EventedObj.Events()
    ) -> 'Frame':
        if style:
            self.setFrameStyle(style)
        if layout:
            self.setLayout(layout)
        return await sizes(await events(self))



from PySide6.QtCore import QSettings as _QSettings
import socket as _socket
from typing import Any


class ContextAPI:
    __settings = _QSettings(_socket.gethostname(), __file__)
    debug: bool = False

    def __getitem__(self, key: str) -> Any:
        return self.__settings.value(key)

    def __setitem__(self, key: str, value: Any) -> None:
        return self.__settings.setValue(key, value)


CONTEXT = ContextAPI()



from PySide6.QtCore import Qt


class Alignment:
    Alignment = Qt.AlignmentFlag

    Left = Qt.AlignLeft
    LeftTop = Qt.AlignLeft | Qt.AlignTop
    LeftCenter = Qt.AlignLeft | Qt.AlignVCenter
    LeftBottom = Qt.AlignLeft | Qt.AlignBottom
    Right = Qt.AlignRight
    RightTop = Qt.AlignRight | Qt.AlignTop
    RightCenter = Qt.AlignRight | Qt.AlignVCenter
    RightBottom = Qt.AlignRight | Qt.AlignBottom
    VCenter = Qt.AlignVCenter
    HCenter = Qt.AlignHCenter
    Top = Qt.AlignTop
    TopCenter = Qt.AlignHCenter | Qt.AlignTop
    Bottom = Qt.AlignBottom
    BottomCenter = Qt.AlignHCenter | Qt.AlignBottom
    Center = Qt.AlignHCenter | Qt.AlignVCenter



from PySide6.QtWidgets import QWidget, QSplitter, QSizePolicy
from PySide6.QtCore import Qt
from typing import Iterable

from .widget import Widget
from .extensions import SplitterWidgetExt
from ..objects import ContextObj
from ..enums import Orientation


class SplitterHandle(Widget):
    def __init__(self, parent: QWidget, name: str):
        super().__init__(parent, name, True)


class SplitterWidget(SplitterWidgetExt, Widget):
    def __init__(self, parent: QWidget, name: str,
                 expand_to: int, expand_min: int = None, expand_max: int = None,
                 orientation: Orientation.Orientation = None):
        Widget.__init__(self, parent, name, True)
        SplitterWidgetExt.__init__(self, expand_to, expand_min, expand_max, orientation)


class Splitter(ContextObj, QSplitter):
    def __init__(self, parent: QWidget, name: str, visible: bool = True, stylesheet: str = None,
                 orientation: Orientation.Orientation = Orientation.Horizontal,
                 policy: tuple[QSizePolicy, QSizePolicy] = (QSizePolicy.Expanding, QSizePolicy.Expanding)):
        QSplitter.__init__(self, orientation, parent)
        ContextObj.__init__(self, parent, name, visible)
        if stylesheet:
            self.setStyleSheet(stylesheet)
            self.setAttribute(Qt.WA_StyledBackground, True)
        self.setSizePolicy(*policy)

    async def init(
            self, *,
            items: Iterable[QWidget] = ()
    ) -> 'Splitter':
        for item in items:
            self.addWidget(item)
        return self

    def addWidget(self, widget: QWidget, collapsible: bool = True) -> None:
        super().addWidget(widget)
        self.setCollapsible(self.count() - 1, collapsible)
        widget.splitter = self



from typing import Any
import aiohttp
import ujson


class Request:
    __base_url = None

    @property
    async def base_url(self) -> str | None:
        if isinstance(self.__base_url, str):
            return self.__base_url
        if self.__base_url.__class__.__name__ == 'function':
            url = self.__base_url()
            if url.__class__.__name__ == 'coroutine':
                return await url
            return url

    @base_url.setter
    def base_url(self, base_url):
        self.__base_url = base_url

    async def __call__(
            self,
            method: str, url: str,
            *,
            params: dict[str, Any] = None, body: dict[str, Any] = None, headers: dict[str, Any] = None,
            data: dict[str, Any] = None,
            nobase: bool = False
    ) -> Any:
        assert (method := method.lower()) in ('get', 'post', 'put', 'delete')
        async with aiohttp.ClientSession(json_serialize=ujson.dumps) as session:
            # convert params values to str
            for key, value in params.items() if params else {}:
                params[key] = str(value)
            async with getattr(session, method)(
                    await self.__url(url, nobase), json=body, params=params, headers=headers, data=data
            ) as response:
                return await response.json()

    async def __url(self, url: str, nobase: bool = False):
        if not (base := await self.base_url) or nobase:
            return url
        if base.endswith('/') and url.startswith('/'):
            return base[:-1] + url
        elif base.endswith('/') and not url.startswith('/'):
            return base + url
        elif not base.endswith('/') and url.startswith('/'):
            return base + url
        elif not base.endswith('/') and not url.startswith('/'):
            return f'{base}/{url}'


request = Request()



from PySide6.QtWidgets import QSlider, QWidget

from ..objects import ContextObj
from ..enums import Orientation


class Slider(ContextObj, Orientation, QSlider):
    def __init__(self, parent: QWidget, name: str, visible: bool = True):
        QSlider.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)
    #
    # async def init(
    #         self, *,
    #         orientation: OrientationExt.Orientation, step: int = 1,
    #         on_change: callable = None
    # ) -> 'Slider':
    #     self.setFocusPolicy(Qt.StrongFocus)
    #     self.setTickPosition(QSlider.TicksBothSides)
    #     self.setTickInterval(10)
    #     self.setSingleStep(10)
    #     self.setValue(self.controls.get(display=display))
    #     if on_change:
    #         self.valueChanged.connect(lambda: self.controls.set(display=display, value=slider.value()))
    #     return self



from .widget import Widget
from .frame import Frame
from .stacked_widget import StackedWidget
from .splitter import Splitter, SplitterHandle, SplitterWidget
from .scroll_area import ScrollArea

from .button import Button
from .label import Label
from .line_input import LineInput
from .text_input import TextInput
from .selector import Selector
from .popup import Popup
from .statusbar import StatusBar
from .window import Window
from .slider import Slider

from .layout import Layout
from .spacer import Spacer

from PySide6.QtWidgets import QWidget
Parent = QWidget



from PySide6.QtGui import QFontMetrics, QPaintEvent
from PySide6.QtWidgets import QLabel, QWidget

from ..objects import ContextObj, SizedObj, EventedObj
from ..enums import Alignment, ElideMode
from ..types import Applicable, Icon


class Label(ContextObj, Alignment, ElideMode, QLabel):
    __elide_mode: ElideMode.ElideMode
    __elide_text: str

    def __init__(self, parent: QWidget, name: str, visible: bool = True):
        QLabel.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)

    async def init(
            self, *,
            text: str = '', wrap: bool = False, icon: Icon = None, elide: ElideMode.ElideMode = None,
            sizes: Applicable = SizedObj.Sizes(), events: Applicable = EventedObj.Events()
    ) -> 'Label':
        self.__elide_mode, self.__elide_text = elide, text
        self.setText(text)
        self.setWordWrap(wrap)
        if icon:
            self.setPixmap(icon.icon.pixmap(icon.size))
        return await sizes(await events(self))

    def paintEvent(self, event: QPaintEvent):
        if self.elided():
            elided_text = self.__elide_text
            self.setText(QFontMetrics(self.font()).elidedText(elided_text, self.__elide_mode, self.width() - 10))
            self.__elide_text = elided_text
        super().paintEvent(event)

    def elided(self) -> bool:
        return self.__elide_mode is not None

    def setText(self, text: str) -> None:
        if self.elided():
            self.__elide_text = text
        super().setText(text)



from PySide6.QtWidgets import QWidget, QCompleter, QStyledItemDelegate
from PySide6.QtCore import Qt
from typing import Sequence

from ..line_input import LineInput
from ...types import Applicable
from ...objects import SizedObj, EventedObj


class SearchBar(LineInput):
    class Completer(QCompleter):
        def __init__(self, parent: QWidget, *, items: Sequence[str], stylesheet: str = ''):
            super().__init__(items, parent)
            self.setCaseSensitivity(Qt.CaseInsensitive)
            self.popup().setObjectName(f'{parent.objectName()}Popup')
            self.popup().setItemDelegate(QStyledItemDelegate(self))
            self.popup().setStyleSheet(stylesheet)

    def __init__(self, parent: QWidget, name: str = None, visible: bool = True):
        LineInput.__init__(self, parent, name if name else self.__class__.__name__, visible)

    async def init(
            self, *,
            placeholder: str, completer: Completer, hidden: bool = False,
            sizes: Applicable = SizedObj.Sizes(), events: Applicable = EventedObj.Events()
    ) -> 'LineInput':
        await super().init(placeholder=placeholder, hidden=hidden, events=events, sizes=sizes)
        self.setCompleter(completer)
        return self



from typing import Callable, Awaitable
import functools
import asyncio
import qasync
import sys
from qasync import asyncSlot  # to import `asyncSlot` from `qasyncio`


class AsyncApp:
    @staticmethod
    def run(amain: Callable[[], Awaitable[None]]) -> None:
        async def wrap():
            def close_future(future: asyncio.Future, loop: asyncio.AbstractEventLoop) -> None:
                loop.call_later(10, future.cancel)
                future.cancel()
            future = asyncio.Future()
            if hasattr(qapp := qasync.QApplication.instance(), 'aboutToQuit'):
                getattr(qapp, 'aboutToQuit').connect(functools.partial(close_future, future, asyncio.get_event_loop()))
            await amain()
            await future

        try:
            qasync.run(wrap())
        except asyncio.exceptions.CancelledError:
            sys.exit(0)


def to_async_slot(to_call):
    @asyncSlot()
    async def wrapped():
        print(type(to_call))
        if to_call.__class__.__name__ in ('function', 'method'):
            await to_call()
        # elif
    return wrapped



from PySide6.QtWidgets import QWidget, QLayout
from PySide6.QtCore import Qt

from ..objects import ContextObj, SizedObj, EventedObj
from ..types import Applicable


class Widget(ContextObj, QWidget):
    def __init__(self, parent: QWidget, name: str, visible: bool = True, stylesheet: str = ''):
        QWidget.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)
        self.setAttribute(Qt.WA_StyledBackground, True)
        self.setStyleSheet(stylesheet)

    async def init(
            self, *,
            layout: QLayout = None,
            sizes: Applicable = SizedObj.Sizes(), events: Applicable = EventedObj.Events()
    ) -> 'Widget':
        if layout:
            self.setLayout(layout)
        return await sizes(await events(self))



from PySide6.QtWidgets import QWidget

from ..button import Button
from ..label import Label
from ..layout import Layout
from ...types import Icon, Size


class MenuButton(Button):
    def __init__(self, parent: QWidget, name: str = None, visible: bool = True):
        super().__init__(parent, name if name else self.__class__.__name__, visible)

    async def init(
            self, *,
            icon: Icon, text: str, total: int = 0, slot: callable = None
    ) -> 'MenuButton':
        self.setLayout(await Layout.horizontal().init(
            margins=(10, 5, 10, 5), spacing=10,
            items=[
                icon_btn := await Button(self, f'{self.objectName()}IconBtn').init(
                    icon=icon, disabled=True, sizes=Button.Sizes(fixed_size=Size(icon.size.width(), icon.size.height()))
                ), Layout.Left,
                text_lbl := await Label(self, f'{self.objectName()}TextLbl').init(
                    text=text, sizes=Label.Sizes(hpolicy=Label.Expanding), elide=Label.ElideRight
                ),
                total_lbl := await Label(self, f'{self.objectName()}TotalLbl').init(
                    text=str(total)
                ), Layout.Right
            ]
        ))
        self.icon_btn = icon_btn
        self.text_lbl = text_lbl
        self.total_lbl = total_lbl
        if slot:
            self.clicked.connect(slot)
        return self



from PySide6.QtWidgets import QTextEdit, QWidget

from ..objects import ContextObj, SizedObj, EventedObj
from ..types import Applicable


class TextInput(ContextObj, QTextEdit):
    def __init__(self, parent: QWidget, name: str, visible: bool = True):
        QTextEdit.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)

    async def init(
            self, *,
            placeholder: str = '', text: str = '',
            sizes: Applicable = SizedObj.Sizes(), events: Applicable = EventedObj.Events()
    ) -> 'TextInput':
        self.setText(text)
        self.setPlaceholderText(placeholder)
        return await sizes(await events(self))



from uuid import uuid4, UUID
from threading import Thread


class ConditionalThread(Thread):
    to_complete: bool = True

    def kill(self):
        self.to_complete = False


class ConditionalThreadQueue:
    def __init__(self):
        self.__threads: dict[UUID, ConditionalThread] = {}

    def new(self, pre: callable, post: callable) -> UUID:
        if self.__threads:  # kill prev thread if exists
            key = list(self.__threads.keys())[-1]
            self.__threads[key].kill()

        def target(thread_uuid: UUID):  # create target for the new thread
            pre()
            thread = self.__threads[thread_uuid]
            if thread.to_complete:  # call post() if thread was not killed
                post()
            self.__threads.pop(thread_uuid)

        uuid = uuid4()
        thread = ConditionalThread(target=target, args=(uuid, ))
        self.__threads[uuid] = thread
        thread.start()
        return uuid

    def __getitem__(self, uuid: UUID):
        return self.__threads[uuid]



from PySide6.QtWidgets import QStatusBar, QWidget

from ..objects import ContextObj


class StatusBar(ContextObj, QStatusBar):
    def __init__(self, parent: QWidget, name: str, visible: bool = True):
        QStatusBar.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)

    async def init(self) -> 'StatusBar':
        return self



class Size:
    def __init__(self, width: int = ..., height: int = ...):
        self.w: int = width if width is not Ellipsis else None
        self.h: int = height if height is not Ellipsis else None

    @property
    def size(self):
        return self.w, self.h



from PySide6.QtWidgets import QWidget, QFileDialog
from PySide6.QtGui import QIcon
from contextlib import suppress

from ..button import Button
from ..popup import Popup
from ...types import Icon, Size
from ...qasyncio import asyncSlot


class ImageButton(Button):
    def __init__(self, parent: QWidget, name: str = None, visible: bool = True):
        super().__init__(parent, name if name else self.__class__.__name__, visible)
        self.image_bytes = None

    async def init(
            self, *,
            icon: Icon, slot: callable = lambda: None, directory: str = ''
    ) -> 'ImageButton':
        await super().init(icon=icon, events=Button.Events(on_click=lambda: self.choose_image(slot, directory)))
        self.RemoveImageBtn = await Button(self, f'{self.objectName()}RemoveImageBtn').init(
            icon=Icon('x-circle.svg', (30, 30)), sizes=Button.Sizes(fixed_size=Size(30, 30)),
            events=Button.Events(
                on_click=lambda: Popup(self.core).display(message=f'Remove icon?', on_success=self.remove_icon)
            )
        )
        self.RemoveImageBtn.move(self.width() - 30, 0)
        self.RemoveImageBtn.setStyleSheet(f'''
            #{self.objectName()}RemoveImageBtn {{background-color: transparent; border-radius: 14px;}}
            #{self.objectName()}RemoveImageBtn:hover {{background-color: rgba(255, 255, 255, 0.3);}}
        ''')
        return self

    @property
    def image_bytes_str(self) -> str | None:
        return str(self.image_bytes) if self.image_bytes else None

    def remove_icon(self):
        super().setIcon(QIcon())
        self.image_bytes = None
        with suppress(AttributeError):
            self.RemoveImageBtn.setVisible(False)

    def setIcon(self, icon: QIcon) -> None:
        self.image_bytes = Icon.bytes(icon, self.iconSize())
        super().setIcon(icon)
        with suppress(AttributeError):
            self.RemoveImageBtn.setVisible(True)

    def setDisabled(self, disabled: bool) -> None:
        with suppress(AttributeError):
            self.RemoveImageBtn.setVisible(not disabled)
        super().setDisabled(disabled)

    def setEnabled(self, enabled: bool) -> None:
        with suppress(AttributeError):
            self.RemoveImageBtn.setVisible(enabled)
        super().setEnabled(enabled)

    @asyncSlot()
    async def choose_image(self, slot: callable = lambda: None, directory: str = ''):
        filepath, _ = QFileDialog.getOpenFileName(self, 'Choose image', directory, 'Images (*.jpg)')
        if filepath:
            with open(filepath, 'rb') as file:
                self.setIcon(Icon(file.read()).icon)
            slot()



from typing import Any, Iterable


async def serializable(dictionary: dict[str, Any], exceptions: Iterable[str] = ()) -> dict[str, Any]:
    def validate(value: Any):
        if isinstance(value, bool):
            return True
        elif value is None:
            return False
        elif isinstance(value, str):
            return len(value)

    return {key: value for key, value in dictionary.items() if validate(value) or key in exceptions}


async def find(data: list[dict[str, Any]], key: Any, value: Any) -> tuple[int, dict[str, Any]] | None:
    for index, item in enumerate(data):
        if item.get(key) == value:
            return index, item



from PySide6.QtWidgets import QMainWindow

from ..objects import SizedObj, EventedObj
from ..enums import WindowHint
from ..types import Applicable


class Window(WindowHint, QMainWindow):
    def __init__(self, name: str, stylesheet: str = ''):
        QMainWindow.__init__(self)
        self.setObjectName(name)
        self.setStyleSheet(stylesheet)

    async def init(
            self, *,
            sizes: Applicable = SizedObj.Sizes(), events: Applicable = EventedObj.Events()
    ) -> 'Window':
        return await sizes(await events(self))



from .alignment import Alignment
from .orientation import Orientation
from .size_policy import SizePolicy
from .window_hint import WindowHint
from .scroll_policy import ScrollPolicy
from .elide_mode import ElideMode



from PySide6.QtGui import QResizeEvent
from PySide6.QtWidgets import QMessageBox, QWidget
from qasync import asyncSlot
from typing import Iterable

from .widget import Widget
from .layout import Layout
from .button import Button
from .label import Label
from .frame import Frame


class Popup(Widget):
    YES = QMessageBox.StandardButton.Yes
    NO = QMessageBox.StandardButton.No
    OK = QMessageBox.StandardButton.Ok
    CANCEL = QMessageBox.StandardButton.Cancel

    stylesheet = f'''
        #Popup {{
            background-color: rgba(255, 255, 255, 0.1);
        }}

        #PopupFrame {{
            background-color: #17212B;
            min-width: 400px;
            min-height: 200px;
            border-radius: 20px;
        }}

        #PopupMessageLbl {{
            color: white;
            font-size: 24px;
            background-color: transparent;
        }}

        #PopupOkBtn,
        #PopupCancelBtn,
        #PopupYesBtn,
        #PopupNoBtn {{
            color: white;
            border: none;
            font-size: 18px; 
            border-radius: 5px;
            min-height: 40px;
        }}

        #PopupOkBtn,
        #PopupYesBtn {{
            background-color: darkgreen;
        }}
    
        #PopupOkBtn:hover,
        #PopupYesBtn:hover {{
            background-color: rgba(0, 255, 0, 0.5);
        }}
        
        #PopupCancelBtn,
        #PopupNoBtn {{
            background-color: rgba(182, 0, 40, 1);
        }}

        #PopupCancelBtn:hover,
        #PopupNoBtn:hover {{
            background-color: rgba(182, 0, 40, 0.5);
        }}
        '''

    def __init__(self, parent: QWidget, name: str = None, stylesheet: str = stylesheet):
        super().__init__(parent, name if name else self.__class__.__name__, True, stylesheet)

    @asyncSlot()
    async def display(
            self, *,
            message: str = '', buttons: Iterable = (),
            on_success: callable = lambda: None, on_failure: callable = lambda: None
    ) -> 'Popup':
        if not buttons:
            buttons = [Popup.YES, Popup.NO]
        btns = []
        if Popup.YES in buttons:
            btns.append(await Button(self, f'{self.objectName()}YesBtn').init(
                text='Yes', events=Button.Events(on_click=lambda: self.slot(on_success))
            ))
        if Popup.NO in buttons:
            btns.append(await Button(self, f'{self.objectName()}NoBtn').init(
                text='No', events=Button.Events(on_click=lambda: self.slot(on_failure))
            ))
        if Popup.OK in buttons:
            btns.append(await Button(self, f'{self.objectName()}OkBtn').init(
                text='Ok', events=Button.Events(on_click=lambda: self.slot(on_success))
            ))
        if Popup.CANCEL in buttons:
            btns.append(await Button(self, f'{self.objectName()}CancelBtn').init(
                text='Cancel', events=Button.Events(on_click=lambda: self.slot(on_failure))
            ))

        self.setLayout(await Layout.vertical().init(
            spacing=10, alignment=Layout.Center, margins=(20, 20, 20, 20),
            items=[
                await Frame(self, f'{self.objectName()}Frame').init(
                    layout=await Layout.vertical().init(
                        spacing=20, margins=(20, 20, 20, 20),
                        items=[
                            await Label(self, f'{self.objectName()}MessageLbl').init(
                                text=message, wrap=True, sizes=Label.Sizes(alignment=Layout.Center)
                            ),
                            await Layout.horizontal().init(
                                spacing=20, items=btns
                            )
                        ]
                    )
                )
            ]
        ))
        return self

    @asyncSlot()
    async def slot(self, slot: callable):
        slot()
        self.deleteLater()

    def resizeEvent(self, event: QResizeEvent) -> None:
        self.resize(self.parent().size())



from PySide6.QtWidgets import QComboBox, QWidget
from typing import Iterable, Any

from ..objects import ContextObj, SizedObj, EventedObj
from ..types import Applicable, Icon


class Selector(ContextObj, QComboBox):
    class Item:
        def __init__(self, text: str, icon: Icon = None, data: Any = None):
            self.params = []
            if icon:
                self.params.append(icon)
            self.params.append(str(text))
            if data:
                self.params.append(data)

    def __init__(self, parent: QWidget, name: str, visible: bool = True):
        QComboBox.__init__(self)
        ContextObj.__init__(self, parent, name, visible)

    async def init(
            self, *,
            items: Iterable[Item | str] = (),
            sizes: Applicable = SizedObj.Sizes(), events: Applicable = EventedObj.Events()
    ) -> 'Selector':
        for item in items:
            if isinstance(item, Selector.Item):
                self.addItem(*item.params)
            elif isinstance(item, str):
                self.addItem(item)
        return await sizes(await events(self))

    def setCurrentText(self, text: Any) -> None:
        super().setCurrentText(str(text))



from PySide6.QtWidgets import QWidget, QLayoutItem, QSpacerItem
from PySide6.QtCore import Qt, QObject
from typing import Sequence, Iterable


class LayoutExt:
    async def init(
            self, *,
            margins: tuple[int, ...] = (0, 0, 0, 0), spacing: int = 0, alignment: Qt.Alignment = None,
            items: Sequence[QObject] = ()
    ) -> 'LayoutExt':
        self.setContentsMargins(*margins)
        self.setSpacing(spacing)
        if alignment:
            self.setAlignment(alignment)
        self.add_items(items)
        return self

    def add(self, obj: QObject, alignment: Qt.AlignmentFlag = None) -> QObject:
        if isinstance(obj, QWidget):
            if alignment:
                self.addWidget(obj, alignment=alignment)
            else:
                self.addWidget(obj)
        elif isinstance(obj, QSpacerItem):
            self.addSpacerItem(obj)
        elif isinstance(obj, QLayoutItem):
            self.addLayout(obj)
        else:
            raise TypeError(f'{obj} has unsupported type {type(obj)} to `add` to {self}')
        return obj

    def add_items(self, items: Sequence[QObject]):
        i = 0
        while i < len(items):
            if i + 1 < len(items) and isinstance(items[i + 1], (Qt.AlignmentFlag, Qt.Alignment)):
                self.add(items[i], items[i + 1])
                i += 1
            else:
                self.add(items[i])
            i += 1

    def clear(self, exceptions: Iterable[QObject] = ()):
        for i in reversed(range(self.count())):
            layout_item = self.itemAt(i).widget()
            if not layout_item:
                layout_item = self.itemAt(i).layout()
            if not layout_item:
                layout_item = self.itemAt(i).spacerItem()
            if layout_item not in exceptions:
                layout_item.setParent(None)



from PySide6.QtWidgets import QSpacerItem

from ..enums import SizePolicy


class Spacer(QSpacerItem):
    def __init__(self, horizontal: bool = False, vertical: bool = False):
        super().__init__(
            0, 0,
            SizePolicy.Expanding if horizontal else SizePolicy.Minimum,
            SizePolicy.Expanding if vertical else SizePolicy.Minimum
        )



from PySide6.QtWidgets import QSizePolicy


class SizePolicy:
    SizePolicy = QSizePolicy

    Expanding = QSizePolicy.Policy.Expanding
    Minimum = QSizePolicy.Policy.Minimum
    Maximum = QSizePolicy.Policy.Maximum



from PySide6.QtWidgets import QWidget
from datetime import datetime, time, date, timezone
from calendar import monthrange
from dateutil import parser

from ..frame import Frame
from ..layout import Layout
from ..selector import Selector
from ..label import Label


class DateTimePicker(Frame):
    now = datetime.now()
    default_format = '%d.%m.%Y %H:%M'

    days = [str(day).zfill(2) for day in range(1, monthrange(now.year, now.month)[1] + 1)]
    months = [str(month).zfill(2) for month in range(1, 12 + 1)]
    years = [str(year) for year in range(now.year, now.year + 10)]

    hours = [str(hour).zfill(2) for hour in range(0, 23 + 1)]
    minutes = [str(minute).zfill(2) for minute in range(0, 59 + 1)]
    seconds = [str(second).zfill(2) for second in range(0, 59 + 1)]

    def __init__(self, parent: QWidget, name: str = None, visible: bool = True):
        super().__init__(parent, name if name else self.__class__.__name__, visible)

    async def init(
            self, *,
            show_date: bool = True, show_time: bool = True,
            margins: tuple[int, ...] = (0, 0, 0, 0), spacing: int = 0, alignment: Layout.Alignment = None
    ) -> 'DateTimePicker':
        self.setLayout(await Layout.vertical().init(
            margins=margins, spacing=spacing, alignment=alignment,
            items=[
                await Frame(self, f'{self.objectName()}DateFrame').init(
                    layout=await Layout.horizontal().init(
                        items=[
                            await Label(self, f'{self.objectName()}DateLbl').init(
                                text='Date:'
                            ),
                            day_selector := await Selector(self, f'{self.objectName()}DaySelector').init(
                                items=self.days
                            ),
                            month_selector := await Selector(self, f'{self.objectName()}MonthSelector').init(
                                items=self.months
                            ),
                            year_selector := await Selector(self, f'{self.objectName()}YearSelector').init(
                                items=self.years
                            )
                        ]
                    )
                ),
                await Frame(self, f'{self.objectName()}TimeFrame').init(
                    layout=await Layout.horizontal().init(
                        items=[
                            await Label(self, f'{self.objectName()}TimeLbl').init(
                                text='Time:'
                            ),
                            hour_selector := await Selector(self, f'{self.objectName()}HourSelector').init(
                                items=self.hours
                            ),
                            minute_selector := await Selector(self, f'{self.objectName()}MinuteSelector').init(
                                items=self.minutes
                            ),
                            second_selector := await Selector(self, f'{self.objectName()}SecondSelector').init(
                                items=self.seconds
                            )
                        ]
                    )
                )
            ]
        ))
        self.day_selector = day_selector
        self.month_selector = month_selector
        self.year_selector = year_selector
        self.hour_selector = hour_selector
        self.minute_selector = minute_selector
        self.second_selector = second_selector
        return self

    def get_datetime(self, tz: bool) -> datetime:
        dt = datetime(day=int(self.day_selector.currentText()),
                      month=int(self.month_selector.currentText()),
                      year=int(self.year_selector.currentText()),
                      hour=int(self.hour_selector.currentText()),
                      minute=int(self.minute_selector.currentText()),
                      second=int(self.second_selector.currentText()),
        )
        if tz:
            dt.replace(tzinfo=timezone.utc)
        else:
            dt.replace(tzinfo=None)
        return dt

    def get_date(self) -> date:
        return date(
            day=int(self.day_selector.currentText()),
            month=int(self.month_selector.currentText()),
            year=int(self.year_selector.currentText())
        )

    def get_time(self) -> time:
        return time(
            hour=int(self.hour_selector.currentText()),
            minute=int(self.minute_selector.currentText()),
            second=int(self.second_selector.currentText())
        )

    def set_date(self, dt: date | datetime | str):
        if isinstance(dt, str):
            dt = self.parse(dt)
        self.day_selector.setCurrentText(str(dt.day).zfill(2))
        self.month_selector.setCurrentText(str(dt.month).zfill(2))
        self.year_selector.setCurrentText(str(dt.year).zfill(2))

    def set_time(self, dt: time | datetime):
        if isinstance(dt, str):
            dt = self.parse(dt)
        self.hour_selector.setCurrentText(str(dt.hour).zfill(2))
        self.minute_selector.setCurrentText(str(dt.minute).zfill(2))
        self.second_selector.setCurrentText(str(dt.second).zfill(2))

    def set_datetime(self, dt: datetime):
        self.set_date(dt)
        self.set_time(dt)

    @staticmethod
    def parse(timestr: str) -> datetime:
        return parser.parse(timestr)



from PySide6.QtCore import Qt


class ScrollPolicy:
    ScrollPolicy = Qt.ScrollBarPolicy

    AlwaysOn = Qt.ScrollBarPolicy.ScrollBarAlwaysOn
    AlwaysOff = Qt.ScrollBarPolicy.ScrollBarAlwaysOff
    WhenNeeded = Qt.ScrollBarPolicy.ScrollBarAsNeeded
    


from PySide6.QtWidgets import QWidget, QStackedWidget
from PySide6.QtCore import Qt, QObject
from typing import Iterable

from ..objects import ContextObj, SizedObj, EventedObj
from ..enums import Alignment, Orientation
from ..types import Applicable


class StackedWidget(ContextObj, Orientation, QStackedWidget):
    def __init__(self, parent: QWidget, name: str, visible: bool = True, stylesheet: str = ''):
        QStackedWidget.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)
        self.setStyleSheet(stylesheet)
        self.setAttribute(Qt.WA_StyledBackground, True)

        # workaround which includes override of `setCurrentIndex`, otherwise problems with `parent()`
        self.addWidget(QWidget(self))

    async def init(
            self, *,
            alignment: Alignment.Alignment = Alignment.HCenter, items: Iterable[QObject] = (),
            sizes: Applicable = SizedObj.Sizes(), events: Applicable = EventedObj.Events()
    ) -> 'StackedWidget':
        self.layout().setAlignment(alignment)
        for item in items:
            self.addWidget(item)
        return await sizes(await events(self))

    def setCurrentIndex(self, index: int) -> None:
        super().setCurrentIndex(index + 1)



from PySide6.QtWidgets import QScrollArea, QWidget
from PySide6.QtCore import QObject
from typing import Sequence, Iterable

from .frame import Frame
from .layout import Layout
from ..objects import ContextObj, SizedObj, EventedObj
from ..enums import Orientation, ScrollPolicy
from ..types import Applicable


class ScrollArea(ContextObj, ScrollPolicy, Orientation, QScrollArea):
    def __init__(self, parent: QWidget, name: str, visible: bool = True, stylesheet: str = None):
        QScrollArea.__init__(self, parent)
        ContextObj.__init__(self, parent, name, visible)
        if stylesheet:
            self.setStyleSheet(stylesheet)

    async def init(
            self, *,
            orientation: Orientation.Orientation,
            hpolicy: ScrollPolicy.ScrollPolicy = ScrollPolicy.WhenNeeded,
            vpolicy: ScrollPolicy.ScrollPolicy = ScrollPolicy.WhenNeeded,

            margins: tuple[int, ...] = (0, 0, 0, 0), spacing: int = 0, alignment: Layout.Alignment = None,
            items: Sequence[QObject] = (),
            sizes: Applicable = SizedObj.Sizes(), events: Applicable = EventedObj.Events()
    ) -> 'ScrollArea':
        self.setHorizontalScrollBarPolicy(hpolicy)
        self.setVerticalScrollBarPolicy(vpolicy)
        self.setWidgetResizable(True)
        frame = Frame(self, f'{self.objectName()}Widget')
        self.setWidget(await frame.init(
            layout=await Layout.oriented(orientation, frame, f'{frame.objectName()}Layout').init(
                margins=margins, spacing=spacing, alignment=alignment, items=items
            )
        ))
        return await sizes(await events(self))

    def addWidget(self, widget: QWidget) -> None:
        self.widget().layout().addWidget(widget)

    def clear(self, exceptions: Iterable[QObject] = ()):
        self.widget().layout().clear(exceptions)



from typing import Any


class StyleSheetParser:
    properties: dict[str, Any] = {}

    def parse(self, stylesheet: str):
        stylesheet = stylesheet.replace('{', '')
        stylesheet = stylesheet.replace('}', '')
        stylesheet = stylesheet.replace(' ', '')
        dicts = stylesheet.split(';')
        dicts = [d for d in dicts if len(d)]
        for d in dicts:
            r = d.split(':')
            self.properties[r[0]] = r[1]

    def __setitem__(self, key, value):
        self.properties[key] = value

    def __getitem__(self, item):
        return self.properties[item]

    def styleSheet(self):
        stylesheet = ''
        for key, value in self.properties.items():
            stylesheet += f'{key}: {value};'
        return stylesheet

# from css_parser import CSSParser
# from css_parser.stylesheets.cssstyledeclaration import CSSStyleDeclaration
#
#
# class StylesheetParser:
#     def __init__(self, style: str):
#         self.style: CSSStyleDeclaration = CSSParser().parseStyle(style)
#
#     def __setitem__(self, key, value):
#         self.style.setProperty(key, value)
#
#     def stylesheet(self):
#         style = ''
#         for key in self.style.keys():
#             style += f'{key}: {self.style.getPropertyValue(key)};'
#         return style
#
#
# def update_inner(old_stylesheet, new_stylesheet):
#     old_style = CSSParser().parseStyle(old_stylesheet)
#     new_style = CSSParser().parseStyle(new_stylesheet)
#     for key in new_style.keys():
#         old_style.setProperty(key, new_style.getPropertyValue(key))
#     style = ''
#     for key in old_style.keys():
#         style += f'{key}: {old_style.getPropertyValue(key)};'
#     return style
#
#
# def parse(raw: str):
#     if len(raw.split('#')) == 1:
#         return raw
#     return raw[raw.find('{') + len('{'):raw.find('}')]
#
#
# def update(button, style):
#     name = button.objectName()
#     defstyle = f'#{name} {{{update_inner(parse(button.styleSheet()), style)}}}'
#     hovstyle = f'#{name}:hover {{background-color: rgba(255, 255, 255, 0.3);}}'
#     return f'{defstyle} {hovstyle}'



from PySide6.QtCore import QObject
from loguru import logger
from contextlib import suppress
from PySide6.QtWidgets import (
    QPushButton, QFrame, QLabel, QLineEdit, QTextEdit, QStackedWidget, QComboBox
)

from ..types import Applicable, Event


class EventedObj:
    @staticmethod
    def Events(
            *,
            on_click: Event = None,
            on_change: Event = None,
            # on_resize: Event = None,
            # on_close: Event = None
    ) -> Applicable:
        async def apply(obj):
            if on_click:
                if isinstance(obj, QPushButton):
                    await EventedObj._connect(obj, 'clicked', on_click)
                elif isinstance(obj, (QLabel, QFrame)):
                    obj.mousePressEvent = lambda event: EventedObj._emit(on_click)
                else:
                    EventedObj._error(obj, 'on_click')

            if on_change:
                if isinstance(obj, (QLineEdit, QTextEdit)):
                    await EventedObj._connect(obj, 'textChanged', on_change)
                elif isinstance(obj, QStackedWidget):
                    await EventedObj._connect(obj, 'currentChanged', on_change)
                elif isinstance(obj, QComboBox):
                    await EventedObj._connect(obj, 'currentTextChanged', on_change)
                else:
                    EventedObj._error(obj, 'on_change')

            # if on_close:
            #     def close(self) -> bool:
            #         print(self)
            #         rv = self.close()
            #         EventedObj.emit(on_close)
            #         return rv
            #     obj.close = close
            #
            # if on_resize:
            #     async def resizeEvent(self, event: QResizeEvent) -> None:
            #         self.resizeEvent(self, event)
            #         await on_resize()
            #     obj.resizeEvent = lambda event: EventedObj.emit(lambda: resizeEvent(event))

            return obj
        return apply

    @staticmethod
    async def _connect(obj: QObject, signalname: str, event: Event):
        signal = getattr(obj, signalname)
        with suppress(Exception):
            signal.disconnect()
        signal.connect(event)

    @staticmethod
    def _emit(event: Event) -> None:
        btn = QPushButton()
        btn.setVisible(False)
        btn.clicked.connect(event)
        btn.click()
        btn.deleteLater()

    @staticmethod
    def _error(obj: QObject, event: str) -> None:
        logger.error(f'event `{event}` is not implemented for `{obj.objectName()}\'s` type: {type(obj)}')



from PySide6.QtWidgets import QWidget
from loguru import logger as _logger
from contextlib import suppress as _suppress
from typing import Iterable
import uuid

from ..contextapi import CONTEXT
from .sized_obj import SizedObj
from .evented_obj import EventedObj


class ContextObj(SizedObj, EventedObj):
    __blacklist: dict[str, list[str]] = {}

    def __init__(self, parent: QWidget = None, name: str = str(uuid.uuid4()), visible: bool = True,
                 stylesheet: str | Iterable[str] = None):
        if parent and name:
            self.setObjectName(name)
            self.__register(parent, name, self)
            self.__register(self, parent.objectName(), parent)
            self.__register(CONTEXT, name, self)

            self.core = parent
            while p := self.core.parent():
                self.core = p
        else:
            attributes = []
            if not parent:
                attributes.append('`parent`')
            if not name:
                attributes.append('`name`')
            if CONTEXT.debug:
                _logger.warning(f"{self.__class__.__name__} not registered in `ContextAPI` since {' and '.join(attributes)} not specified")

        with _suppress(Exception):
            self.setVisible(visible)
        if stylesheet:
            with _suppress(Exception):
                if not isinstance(stylesheet, str):
                    stylesheet = ''.join(stylesheet)
                self.setStyleSheet(stylesheet)

    def __register(self, parent: object, name: str, child: QWidget) -> None:
        """
        Check if object not in `blacklist`, then register, otherwise show warning
        Check if object already registered, push its name to `blacklist` unregister and show warning otherwise register
        :param parent:
        :param name:
        :param child:
        :return:
        """
        key = parent.__class__.__name__
        self.__blacklist[key] = self.__blacklist.get(key, [])
        if name not in self.__blacklist[key]:
            try:
                getattr(parent, name)
                delattr(parent, name)
                self.__blacklist[key].append(name)
                if CONTEXT.debug:
                    _logger.warning(f'{name} already registered in {key}')
            except AttributeError:
                setattr(parent, name, child)



from PySide6.QtCore import Qt


class ElideMode:
    ElideMode = Qt.TextElideMode

    ElideRight = Qt.TextElideMode.ElideRight
    ElideLeft = Qt.TextElideMode.ElideLeft
    ElideMiddle = Qt.TextElideMode.ElideMiddle
